#!/usr/bin/env python
import logging
import os
import sys
from argparse import ArgumentParser
from pathlib import Path

from pypromice.pipeline.aws import AWS
from pypromice.io.write import prepare_and_write


def parse_arguments_l2():
    parser = ArgumentParser(description="AWS L2 processor")

    parser.add_argument('-c', '--config_file', type=str, required=True,
                        help='Path to config (TOML) file')
    parser.add_argument('-i', '--inpath', type=str, required=True, 
                        help='Path to input data')
    parser.add_argument('-o', '--outpath', default=None, type=str, required=False, 
                        help='Path where to write output')
    parser.add_argument('-v', '--variables', default=None, type=str, 
                        required=False, help='File path to variables look-up table')
    parser.add_argument('-m', '--metadata', default=None, type=str, 
                        required=False, help='File path to metadata')
    parser.add_argument('--data_issues_path', '--issues', default=None, help="Path to data issues repository")
    args = parser.parse_args()
    return args


def get_l2(config_file, inpath, outpath, variables, metadata, data_issues_path: Path) -> AWS:
    # Define input path
    station_name = config_file.split('/')[-1].split('.')[0] 
    station_path = os.path.join(inpath, station_name)
    
    # checking that data_issues_path is valid
    if data_issues_path is None:
        data_issues_path = Path("../PROMICE-AWS-data-issues")
        if data_issues_path.exists():
            logging.warning(f"data_issues_path is missing. Using default data issues path: {data_issues_path}")
        else:
            raise ValueError("data_issues_path is missing. Please provide a valid path to the data issues repository")

    if os.path.exists(station_path):
        aws = AWS(config_file, 
                  station_path,
                  data_issues_repository=data_issues_path, 
                  var_file=variables, 
                  meta_file=metadata)
    else:
        aws = AWS(config_file, 
                  inpath, 
                  data_issues_repository=data_issues_path, 
                  var_file=variables, 
                  meta_file=metadata)

    # Perform level 1 and 2 processing
    aws.getL1()
    aws.getL2()
    # Write out level 2
    if outpath is not None:
        if not os.path.isdir(outpath):
            os.mkdir(outpath)
        if aws.L2.attrs['format'] == 'raw':
            prepare_and_write(aws.L2, outpath, aws.vars, aws.meta, '10min')
        prepare_and_write(aws.L2, outpath, aws.vars, aws.meta, '60min')
    return aws


def main():
    args = parse_arguments_l2()

    logging.basicConfig(
        format="%(asctime)s; %(levelname)s; %(name)s; %(message)s",
        level=logging.INFO,
        stream=sys.stdout,
    )

    _ = get_l2(
        args.config_file,
        args.inpath,
        args.outpath,
        args.variables,
        args.metadata,
        args.data_issues_path,
    )


if __name__ == "__main__":  
    main()
        


__all__ = ["convert_and_filter_tilt", "apply_tilt_factor",
           "smooth_tilt_with_moving_window", "interpolate_tilt",
           "interpolate_rotation", "calculate_spherical_tilt",
           "calculate_declination", "calculate_hour_angle",
           "calculate_sun_direction_degrees", "calculate_zenith",
           "calculate_angle_difference"]

import numpy as np
import xarray as xr
import pandas as pd

tilt_smoothing_win_size = 7 # Station tilt smoothing window size
tilt_stddev_threshold = 0.2 # Station tilt interpolation standard deviation threshold
rot_stddev_threshold = 4    # Station rotation interpolation standard deviation threshold
tilt_threshold = -100       # Station tilt threshold
deg2rad = np.pi / 180       # Degrees to radians conversion
rad2deg = 1 / deg2rad       # Radians to degrees conversion


def apply_tilt_factor(tilt: xr.DataArray,
                      tilt_correction_factor: float
) -> xr.DataArray:
    """Apply tilt correction factor to station tilt values

    Parameters
    ----------
    tilt : xr.DataArray
        Tilt array (either 'tilt_x' or 'tilt_y')
    tilt_correction_factor : float
        Correction factor to apply to tilt measurements

    Returns
    -------
    xr.DataArray
        Corrected tilt measurements
    """
    return tilt * tilt_correction_factor


def convert_and_filter_tilt(tilt: xr.DataArray
) -> xr.DataArray:
    """Convert station tilt from voltage to degrees,
    and filter tilt with given threshold. Voltage-to-degrees
    conversion is based on the equation in 3.2.9 in Fausto
    et al. (2021) https://doi.org/10.5194/essd-13-3819-2021

    Parameters
    ----------
    tilt : xr.DataArray
        Array (either 'tilt_x' or 'tilt_y'), tilt values (voltage)

    Returns
    -------
    xr.DataArray
        Array (either 'tilt_x' or 'tilt_y'), tilt values (degrees)
    """
    # IDL version:
    # notOKtiltX = where(tiltX lt -100, complement=OKtiltX) & notOKtiltY = where(tiltY lt -100, complement=OKtiltY)
    # tiltX = tiltX/10.
    # tiltnonzero = where(tiltX ne 0 and tiltX gt -40 and tiltX lt 40)
    # if n_elements(tiltnonzero) ne 1 then tiltX[tiltnonzero] = tiltX[tiltnonzero]/abs(tiltX[tiltnonzero])*(-0.49*(abs(tiltX[tiltnonzero]))^4 + 3.6*(abs(tiltX[tiltnonzero]))^3 - 10.4*(abs(tiltX[tiltnonzero]))^2 +21.1*(abs(tiltX[tiltnonzero])))
    # tiltY = tiltY/10.
    # tiltnonzero = where(tiltY ne 0 and tiltY gt -40 and tiltY lt 40)
    # if n_elements(tiltnonzero) ne 1 then tiltY[tiltnonzero] = tiltY[tiltnonzero]/abs(tiltY[tiltnonzero])*(-0.49*(abs(tiltY[tiltnonzero]))^4 + 3.6*(abs(tiltY[tiltnonzero]))^3 - 10.4*(abs(tiltY[tiltnonzero]))^2 +21.1*(abs(tiltY[tiltnonzero])))

    # if n_elements(OKtiltX) gt 1 then tiltX[notOKtiltX] = interpol(tiltX[OKtiltX],OKtiltX,notOKtiltX) ; Interpolate over gaps for radiation correction; set to -999 again below.

    # Define valid tilt values and create mask
    notOKtilt = (tilt < tilt_threshold)
    OKtilt = (tilt >= tilt_threshold)

    # Convert tilt values
    dst = tilt / 10
    nz = (dst != 0) & (np.abs(dst) < 40)
    dst = dst.where(~nz, other = dst / np.abs(dst)
                      * (-0.49
                         * (np.abs(dst))**4 + 3.6
                         * (np.abs(dst))**3 - 10.4
                         * (np.abs(dst))**2 + 21.1
                         * (np.abs(dst))))

    # Apply filtering mask
    dst = dst.where(~notOKtilt)

    # TODO: Filling w/o considering time gaps to re-create IDL/GDL outputs.
    #  Should fill with coordinate not False. Also consider 'max_gap' option?
    return dst.interpolate_na(dim='time', use_coordinate=False)


def smooth_tilt_with_moving_window(tilt: xr.DataArray
) -> tuple[str, np.ndarray]:
    """Smooth tilt values using the pandas 'rolling' window method
    e.g. a value of 7 spans 70 minutes using 10-minute data.
    This is translated from the previous IDL/GDL smoothing algorithm:
    tiltX = smooth(tiltX,7,/EDGE_MIRROR,MISSING=-999) & tiltY = smooth(tiltY,7,/EDGE_MIRROR, MISSING=-999)
    endif
    In Python, this should be
    dstxy = dstxy.rolling(time=7, win_type='boxcar', center=True).mean()
    But the EDGE_MIRROR makes it a bit more complicated

    Parameters
    ----------
    tilt : xr.DataArray
        Array (either 'tilt_x' or 'tilt_y'), tilt values (can be in degrees or voltage)

    Returns
    -------
    tdf_rolling : tuple, as: (str, numpy.ndarray)
        The numpy array is the tilt values, smoothed with a rolling mean
    """
    s = int(tilt_smoothing_win_size/2)
    tdf = tilt.to_dataframe()
    mirror_start = tdf.iloc[:s][::-1]
    mirror_end = tdf.iloc[-s:][::-1]
    mirrored_tdf = pd.concat([mirror_start, tdf, mirror_end])

    tdf_rolling = (
        ("time"),
        mirrored_tdf.rolling(
            tilt_smoothing_win_size, win_type="boxcar", min_periods=1, center=True
            ).mean()[s:-s].values.flatten()
        )
    return tdf_rolling


def interpolate_tilt(tilt: xr.DataArray,
) -> xr.DataArray:
    """Interpolate (fill data gaps) and smooth station tilt
    using a moving standard deviation over a 3-day sliding
    window.

    Parameters
    ----------
    tilt : xr.DataArray
        Either X or Y tilt inclinometer measurements

    Returns
    -------
    xr.DataArray
        Either X or Y smoothed tilt inclinometer measurements
    """
    # We calculate the moving standard deviation over a 3-day sliding window
    # hourly resampling is necessary to make sure the same threshold can be used
    # for 10 min and hourly data
    moving_std_gap_filled = tilt.to_series().resample("h").median().rolling(
                    3*24, center=True, min_periods=2
                    ).std().reindex(tilt.time, method="bfill").values

    # We select the good timestamps and gapfill assuming that
    # - when tilt goes missing the last available value is used
    # - when tilt is not available for the very first time steps, the first
    #   good value is used for backfill
    return tilt.where(
                moving_std_gap_filled < tilt_stddev_threshold
                ).ffill(dim="time").bfill(dim="time")


def interpolate_rotation(rot: xr.DataArray,
                         threshold=4):
    """Interpolate and smooth station rotation

    Parameters
    ----------
    rot : xr.DataArray
        Rotation measurements from inclinometer
    threshold : float
        threshold used in a standard deviation based filter

    Returns
    -------
    xr.DataArray
        smoothed rotation measurements from inclinometer
    """
    moving_std_gap_filled = rot.to_series().resample("h").median().rolling(
                    3*24, center=True, min_periods=2
                    ).std().reindex(rot.time, method="bfill").values

    # Same as for interpolate_tilt() with, in addition:
    #     - a resampling to daily values
    #     - a two-week median smoothing
    #     - a resampling from these daily values to the original temporal resolution
    return ("time", (rot.where(moving_std_gap_filled < rot_stddev_threshold)
            .ffill(dim="time")
            .to_series().resample("D").median()
            .rolling(7*2,center=True,min_periods=2).median()
            .reindex(rot.time, method="bfill").values
            ))


def calculate_spherical_tilt(
    tilt_x: xr.DataArray,
    tilt_y: xr.DataArray
) -> tuple[xr.DataArray, xr.DataArray]:
    """Calculate station tilt

    Parameters
    ----------
    tilt_x : xr.DataArray
        X tilt inclinometer measurements
    tilt_y : xr.DataArray
        Y tilt inclinometer measurements

    Returns
    -------
    phi_sensor_rad : xr.DataArray
        Spherical tilt coordinates
    theta_sensor_rad : xr.DataArray
        Total tilt of sensor, where 0 is horizontal
    """
    # Tilt as radians
    tx = tilt_x * deg2rad
    ty = tilt_y * deg2rad

    # Calculate cartesian coordinates
    X = np.sin(tx) * np.cos(tx) * np.sin(ty)**2 + np.sin(tx) * np.cos(ty)**2
    Y = np.sin(ty) * np.cos(ty) * np.sin(tx)**2 + np.sin(ty) * np.cos(tx)**2
    Z = np.cos(tx) * np.cos(ty) + np.sin(tx)**2 * np.sin(ty)**2

    # Calculate spherical coordinates
    phi_sensor_rad = -np.pi /2 - np.arctan(Y/X)
    phi_sensor_rad[X > 0] += np.pi
    phi_sensor_rad[(X == 0) & (Y < 0)] = np.pi
    phi_sensor_rad[(X == 0) & (Y == 0)] = 0
    phi_sensor_rad[phi_sensor_rad < 0] += 2*np.pi

    # Total tilt of the sensor, i.e. 0 when horizontal
    theta_sensor_rad = np.arccos(Z / (X**2 + Y**2 + Z**2)**0.5)
    return phi_sensor_rad, theta_sensor_rad


def calculate_declination(doy: xr.DataArray,
                          hour: xr.DataArray,
                          minute: xr.DataArray
) -> xr.DataArray:
    """Calculate sun declination based on time

    Parameters
    ----------
    doy : xr.DataArray
        Day of year
    hour : xr.DataArray
        Hour of day
    minute : xr.DataArray
        Minute of hour

    Returns
    -------
    xr.DataArray
        Sun declination in radians
    """
    d0_rad = 2 * np.pi * (doy + (hour + minute / 60) / 24 -1) / 365
    return np.arcsin(0.006918 - 0.399912
                     * np.cos(d0_rad) + 0.070257
                     * np.sin(d0_rad) - 0.006758
                     * np.cos(2 * d0_rad) + 0.000907
                     * np.sin(2 * d0_rad) - 0.002697
                     * np.cos(3 * d0_rad) + 0.00148
                     * np.sin(3 * d0_rad))


def calculate_hour_angle(hour: xr.DataArray,
                         minute: xr.DataArray,
                         lon: float
) -> xr.DataArray:
    """Calculate hour angle of sun based on time and longitude. Make sure that
    time is set to UTC and longitude is positive when west. Hour angle should
    be 0 at noon

    Parameters
    ----------
    hour : xr.DataArray
        Hour of day
    minute : xr.DataArray
        Minute of hour
    lon : float
        Longitude

    Returns
    -------
    xr.DataArray
        Hour angle of sun
    """
    return 2 * np.pi * (((hour + minute / 60) / 24 - 0.5) - lon/360)
     # ; - 15.*timezone/360.)


def calculate_sun_direction_degrees(HourAngle_rad: xr.DataArray
) -> xr.DataArray:
    """Calculate sun direction as degrees. This is an alternative to
    calculate_hour_angle that is currently not implemented into the official
    L0>>L3 workflow. Here, 180 degrees is at noon (NH), as opposed to
    HourAngle

    Parameters
    ----------
    HourAngle_rad : xr.DataArray
        Sun hour angle in radians

    Returns
    -------
    DirectionSun_deg : xr.DataArray
        Sun direction in degrees
    """
    DirectionSun_deg = HourAngle_rad * 180/np.pi - 180
    DirectionSun_deg[DirectionSun_deg < 0] += 360
    DirectionSun_deg[DirectionSun_deg < 0] += 360
    return DirectionSun_deg


def calculate_zenith(lat: float,
                     Declination_rad: xr.DataArray,
                     HourAngle_rad: xr.DataArray
) -> tuple[xr.DataArray, xr.DataArray]:
    """Calculate sun zenith in radians and degrees

    Parameters
    ----------
    lat : float
        Latitude
    Declination_rad : xr.DataArray
        Sun declination in radians
    HourAngle_rad : xr.DataArray
        Sun hour angle in radians

    Returns
    -------
    ZenithAngle_rad : xr.DataArray
        Zenith angle in radians
    ZenithAngle_deg : xr.DataArray
        Zenith angle in degrees
"""
    ZenithAngle_rad = np.arccos(np.cos(lat * deg2rad)
                                * np.cos(Declination_rad)
                                * np.cos(HourAngle_rad)
                                + np.sin(lat * deg2rad)
                                * np.sin(Declination_rad))

    ZenithAngle_deg = ZenithAngle_rad * rad2deg
    return ZenithAngle_rad, ZenithAngle_deg


def calculate_angle_difference(ZenithAngle_rad: xr.DataArray,
                               HourAngle_rad: xr.DataArray,
                               phi_sensor_rad: xr.DataArray,
                               theta_sensor_rad: xr.DataArray
) -> xr.DataArray:
    """Calculate angle between sun and upper sensor (to determine when sun is
    in sight of upper radiometer sensor)

    Parameters
    ----------
    ZenithAngle_rad : xr.DataArray
        Zenith angle in radians
    HourAngle_rad : xr.DataArray
        Sun hour angle in radians
    phi_sensor_rad : xarray.DataArray
        Spherical tilt coordinates
    theta_sensor_rad : xarray.DataArray
        Total tilt of sensor, where 0 is horizontal

    Returns
    -------
    xr.DataArray
        Angle between sun and sensor
    """
    return 180 / np.pi * np.arccos(np.sin(ZenithAngle_rad)
                                   * np.cos(HourAngle_rad + np.pi)
                                   * np.sin(theta_sensor_rad)
                                   * np.cos(phi_sensor_rad)
                                   + np.sin(ZenithAngle_rad)
                                   * np.sin(HourAngle_rad + np.pi)
                                   * np.sin(theta_sensor_rad)
                                   * np.sin(phi_sensor_rad)
                                   + np.cos(ZenithAngle_rad)
                                   * np.cos(theta_sensor_rad))


#!/usr/bin/env python
import email
import imaplib
import os
from argparse import ArgumentParser
from configparser import ConfigParser
from glob import glob

from pypromice.tx.tx import getMail


def parse_arguments_msg():
    parser = ArgumentParser(description="AWS message downloader")       
    parser.add_argument('-a', '--account', default=None, type=str, 
                        required=True, help='Email account .ini file')
    parser.add_argument('-p', '--password', default=None, type=str, 
                        required=True, help='Email credentials .ini file')                      
    parser.add_argument('-o', '--outpath', default='.', type=str, 
                        required=False, help='Path where to write output')    	
    parser.add_argument('-u', '--uid', default=None, type=str, 
                        required=False, help='Last uid, defined from files if not given')	
    parser.add_argument('-m', '--mailbox', default="'[Gmail]/All Mail'",
                        type=str, required=False, help='Mailbox folder to collect messages from')        
    args = parser.parse_args()
    return args


def get_msg():
    args = parse_arguments_msg()

    # Set credential paths
    accounts_file = args.account
    credentials_file = args.password

    # Set last aws uid path
    if args.uid is not None:
        last_uid = int(args.uid)
    else:
        last_uid = None

    # Set output file directory
    out_dir = args.outpath

    # Define accounts and credentials ini file paths
    accounts_ini = ConfigParser()
    accounts_ini.read_file(open(accounts_file))
    accounts_ini.read(credentials_file)

    # Get credentials
    account = accounts_ini.get("aws", "account")
    server = accounts_ini.get("aws", "server")
    port = accounts_ini.getint("aws", "port")
    password = accounts_ini.get("aws", "password")
    if not password:
        password = input("password for AWS email account: ")
    print("AWS data from server %s, account %s" % (server, account))

    # Define last UID from file names
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)
        if last_uid is None:
            last_uid = 0
    else:
        if last_uid is None:
            fl = list(glob(os.path.join(out_dir, "*.msg")))
            fn = [int(f.split("_")[-1].split(".")[0]) for f in fl]
            last_uid = max(fn)

    print("Fetching mail from uid %i" % last_uid)

    # Log in to email server
    mail_server = imaplib.IMAP4_SSL(server, port)
    typ, accountDetails = mail_server.login(account, password)
    if typ != "OK":
        print("Not able to sign in!")
        raise

    # Grab new emails
    result, data = mail_server.select(mailbox=args.mailbox, readonly=True)
    print('mailbox contains %s messages' %data[0])


    # Save mail to file
    for uid, mail in getMail(mail_server, last_uid=last_uid):
        message = email.message_from_string(mail)
        outfile = str(message["subject"]) + "_" + str(uid) + ".msg"

        with open(os.path.join(out_dir, outfile), "w") as out:
            out.write(mail)

    # Close mail server if open
    if "mail_server" in locals():
        print(f"\nClosing {account}")
        mail_server.close()
        resp = mail_server.logout()
        assert resp[0].upper() == "BYE"

    print("Finished")


if __name__ == "__main__":
    get_msg()



__all__=['correct_wind_speed', 'filter_wind_direction', 'calculate_directional_wind_speed']

import numpy as np
import xarray as xr

DEG2RAD=np.pi/180

def correct_wind_speed(wspd: xr.DataArray, coefficient) -> xr.DataArray:
    """Correct wind speed with a linear correction coefficient. For example, the conversion from a standard
    Young anemometer to an Arctic Young anemometer is 1.7.

    Parameters
    ----------
    wspd : xr.DataArray
        Wind speed
    coefficient : float
        Correction coefficient

    Returns
    -------
    xr.DataArray
        Corrected wind speed
    """
    return wspd * coefficient

def filter_wind_direction(wdir: xr.DataArray, wspd: xr.DataArray) -> xr.DataArray:
    """Filter wind direction by wind speed, where wind direction values are removed if
    wind speed is zero.

    Parameters
    ----------
    wdir : xr.DataArray
        Wind direction
    wspd : xr.DataArray
        Wind speed

    Returns
    -------
    xr.DataArray
        Filtered wind direction
    """
    return wdir.where(wspd != 0)


def calculate_directional_wind_speed(wspd: xr.DataArray, wdir: xr.DataArray):
    """Calculate directional wind speed from wind speed and direction

    Parameters
    ----------
    wspd : xr.DataArray
        Wind speed data array
    wdir : xr.DataArray
        Wind direction data array
    deg2rad : float
        Degree to radians coefficient. The default is np.pi/180

    Returns
    -------
    wspd_x : xr.DataArray
        Wind speed in X direction
    wspd_y : xr.DatArray
        Wind speed in Y direction
    """
    wspd_x = wspd * np.sin(wdir * DEG2RAD)
    wspd_y = wspd * np.cos(wdir * DEG2RAD)
    return wspd_x, wspd_y


#!/usr/bin/env python

import logging
import os
import re
import sys
import time
from argparse import ArgumentParser
from collections import deque
from datetime import datetime, timedelta
from glob import glob

import toml

from pypromice.resources import DEFAULT_PAYLOAD_FORMATS_PATH, DEFAULT_PAYLOAD_TYPES_PATH
from pypromice.tx.email_client.base_mail_client import BaseMailClient
from pypromice.tx.email_client.imap_client import IMAPClient
from pypromice.tx.tx import L0tx


logger = logging.getLogger(__name__)


def parse_arguments_l0tx():
    parser = ArgumentParser(description="AWS L0 transmission fetcher")
    parser.add_argument('--account', '-a', type=str, required=True, help='Email account .ini file')
    parser.add_argument('--password', '-p', type=str, required=True, help='Email credentials .ini file')
    parser.add_argument('--config', '-c', type=str, required=True, help='Directory to config .toml files')
    parser.add_argument('--uid', '-u', type=str, required=True, help='Last AWS uid .ini file')

    parser.add_argument('--outpath', '-o', default=None, type=str, required=False, help='Path where to write output (if given)')
    parser.add_argument('--awsname', '-n', default='*', type=str, required=False, help='name of the AWS to be fetched')
    parser.add_argument('--formats', '-f',default=DEFAULT_PAYLOAD_FORMATS_PATH, type=str, required=False, help='Path to Payload format .csv file')
    parser.add_argument('--types', '-t',default=DEFAULT_PAYLOAD_TYPES_PATH, type=str, required=False, help='Path to Payload type .csv file')
    
    args = parser.parse_args()
    return args


def get_l0tx():
    args = parse_arguments_l0tx()
    toml_path = os.path.join(args.config, args.awsname + ".toml")
    toml_list = glob(toml_path)

    aws_modem_configurations = {}

    for t in toml_list:
        conf = toml.load(t)
        count = 1
        for m in conf["modem"]:
            name = str(conf["station_id"]) + "_" + m[0] + "_" + str(count)

            if len(m[1:]) == 1:
                aws_modem_configurations[name] = [
                    datetime.strptime(m[1], "%Y-%m-%d %H:%M:%S"),
                    datetime.now() + timedelta(hours=3),
                ]
            elif len(m[1:]) == 2:
                aws_modem_configurations[name] = [
                    datetime.strptime(m[1], "%Y-%m-%d %H:%M:%S"),
                    datetime.strptime(m[2], "%Y-%m-%d %H:%M:%S"),
                ]
            count += 1

    # ----------------------------------
    # Set payload formatter paths
    formatter_file = args.formats
    type_file = args.types
    # Set credential paths
    accounts_file = args.account
    credentials_file = args.password
    # Set last aws uid path
    uid_file = args.uid

    aws_name = args.awsname

    # Set output file directory
    out_dir = args.outpath
    if out_dir is not None:
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

    # ----------------------------------
    
    # Get last email uid
    with open(uid_file, "r") as last_uid_f:
        last_uid = int(last_uid_f.readline())

    # ----------------------------------
    with IMAPClient.from_config_files(
        accounts_file,
        credentials_file,
    ) as mail_client:
        uid = fetch_transmission_data(
            uid=last_uid,
            mail_client=mail_client,
            out_dir=out_dir,
            formatter_file=formatter_file,
            type_file=type_file,
            aws_modem_configurations=aws_modem_configurations,
        )

    finalize_output_files(aws_name, out_dir)

    # Write last aws uid to ini file
    # ----------------------------------
    try:
        with open(uid_file, "w") as last_uid_f:
            last_uid_f.write(str(uid))
    except:
        logger.warning(f'Could not write last uid {uid} to {uid_file}')

    logger.info("Finished")


def fetch_transmission_data(
        uid,
        mail_client: BaseMailClient,
        out_dir,
        formatter_file,
        type_file,
        aws_modem_configurations,
):
    """
    Fetches and processes messages from a mail inbox since a given UID, extracts relevant
    information based on the specified AWS modem configurations, and logs or stores the
    processed data in an output directory.

    This function iterates over messages received since the specified UID, parses the subject
    and date, checks for matching IMEI in the provided AWS modem configurations, validates
    date constraints, and processes the message content accordingly. If the conditions are
    met, it logs the message details and writes processed data to an output file if
    configured.

    Parameters
    ----------
    uid : int
        The unique identifier of the last processed email to start fetching messages from.
    mail_client : BaseMailClient
        An instance of a mail client providing access to the inbox and email messages.
    out_dir : str or None
        The path to the directory where processed messages will be written. If None, messages
        are not written to disk.
    formatter_file : Any
        The formatter file used to process the message content.
    type_file : Any
        The type file used to categorize or handle message processing.
    aws_modem_configurations : dict
        A mapping of IMEI keys to datetime range values, determining which messages to
        process based on their content and the timestamp in the AWS modem configurations.

    Returns
    -------
    int
        Returns the UID of the last processed message.
    """
    for uid, message in mail_client.iter_messages_since(uid):

        try:
            (imei,) = re.findall(r"[0-9]+", message.get_all("subject")[0])
            d = datetime.strptime(
                message.get_all("date")[0], "%a, %d %b %Y %H:%M:%S %Z"
            )
        except:
            imei = None
            d = None

        for k, v in aws_modem_configurations.items():
            if str(imei) in k:
                if v[0] < d < v[1]:
                    logger.info(
                        f'AWS message for {k}.txt, {d.strftime("%Y-%m-%d %H:%M:%S")}'
                    )
                    l0 = L0tx(message, formatter_file, type_file)
                    
                    if l0.msg:
                        body = l0.getEmailBody()
                        if "Lat" in body[0] and "Long" in body[0]:
                            lat_idx = body[0].split().index("Lat") + 2
                            lon_idx = lat_pos = body[0].split().index("Long") + 2
                            lat = body[0].split()[lat_idx]
                            lon = body[0].split()[lon_idx]
                            # append lat and lon to end of message
                            l0.msg = l0.msg + ",{},{}".format(lat, lon)
                        logger.info(l0.msg)

                        if out_dir is not None:
                            out_fn = str(k) + str(l0.flag) + ".txt"
                            out_path = os.sep.join((out_dir, out_fn))
                            logger.info(f"Writing to {out_fn}")
                            with open(out_path, mode="a") as out_f:
                                out_f.write(l0.msg + "\n")

    return uid


def finalize_output_files(aws_name, out_dir):
    """
    Finalize output files by sorting and deduplicating lines in text files.

    This function processes text files in the specified output directory that match
    the given AWS name and have been modified within the past hour. It sorts the
    lines in these files, removes duplicates, and generates new sorted files.

    Parameters
    ----------
    aws_name : str
        The prefix of the file names to process, typically related to the AWS instance name.
    out_dir : str or None
        The directory path where the files are located. If None, the function exits
        without processing any files.

    Notes
    -----
    - It only processes files matching the AWS name and `.txt` extension.
    - Modified files are identified based on a time window of the past hour.
    - The function appends "sorted_" to the filenames of the output files, which will
      contain sorted and deduplicated lines.
    """
    # ----------------------------------
    if out_dir is not None:

        # Find modified files (within the last hour)
        mfiles = [
            mfile
            for mfile in glob(out_dir + "/" + aws_name + "*.txt")

            if isModified(mfile, 1)
        ]

        # Sort L0tx files and add tails
        for f in mfiles:
            # Sort lines in L0tx file and remove duplicates
            in_dirn, in_fn = os.path.split(f)
            out_fn = "sorted_" + in_fn
            out_pn = os.sep.join((in_dirn, out_fn))
            sortLines(f, out_pn)


def findDuplicates(lines):
    """Find duplicates lines in list of strings

    Parameters
    ---------
    lines : list
       List of strings

    Returns
    -------
    unique_lines : list
       List of unique strings
    """
    unique_lines = list(set(lines))
    duplicates_count = len(lines) - len(unique_lines)
    logger.info(f'{duplicates_count} duplicates found')
    return unique_lines


def sortLines(in_file, out_file, replace_unsorted=True):                       #Formerly called sorter.py
    """Sort lines in text file

    Parameters
    ----------
    in_file : str
        Input file path
    out_file : str
        Output file path
    replace_unsorted : bool, optional
        Flag to replace unsorted files with sorted files. The default is True.
    """
    logger.info(f'\nSorting {in_file}')

    # Open input file and read lines
    with open(in_file) as in_f:
        lines = in_f.readlines()

    # Remove duplicate lines and sort
    unique_lines = findDuplicates(lines.copy())
    unique_lines.sort()
    if lines != unique_lines:
        # Write sorted file
        with open(out_file, 'w') as out_f:
            # out_f.write(headers)
            out_f.writelines(unique_lines)

        # Replace input file with new sorted file
        if replace_unsorted:
            os.remove(in_file)
            os.rename(out_file, in_file)


def addTail(in_file, out_dir, aws_name, header_names='', lines_limit=100):
    """Generate tails file from L0tx file

    Parameters
    ----------
    in_file : str
        Input L0tx file
    out_dir : str
        Output directory for tails file
    aws_name : str
        AWS name
    header_names : str, optional
        Header names. The default is ''.
    lines_limit : int, optional
        Number of lines to append to tails file. The default is 100.
    """
    with open(in_file) as in_f:
        tail = deque(in_f, lines_limit)

    headers_lines = [l + '\n' for l in header_names.split('\n')]
    tail = list(set(headers_lines + list(tail)))
    tail.sort()

    out_fn = '_'.join((aws_name, in_file.split('/')[-1]))
    out_pn = os.sep.join((out_dir, out_fn))

    with open(out_pn, 'w') as out_f:
        #out_f.write(headers)
        out_f.writelines(tail)
        logger.info(f'Tails file written to {out_pn}')


def isModified(filename, time_threshold=1):
    """Return flag denoting if file is modified within a certain timeframe

    Parameters
    ----------
    filename : str
        File path
    time_threshold : int
        Time threshold (provided in hours)

    Returns
    -------
    bool
        Flag denoting if modified (True) or not (False)
    """
    delta = time.time() - os.path.getmtime(filename)
    delta = delta / (60*60)
    if delta < time_threshold:
        return True
    return False


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s; %(levelname)s; %(name)s; %(message)s",
        stream=sys.stdout,
    )
    get_l0tx()



"""
Module for decoding payloads from a data logging system.

This module provides functions for decoding binary payloads into meaningful
data values based on specified binary formats. It includes specific decoders
for two- and four-byte encodings, as well as functionality for determining
payload formats and decoding entire payloads. A custom exception, `DecodeError`,
is implemented to handle errors that may occur during the decoding process.
Additionally, logging is used for debug information throughout the decoding
process.
"""
import logging
import numpy as np
from datetime import datetime
from pathlib import Path

__all__ = [
    "DecodeError",
    "determine_payload_format",
    "decode"
]

logger = logging.getLogger(__name__)

CR_BASIC_EPOCH_OFFSET = datetime(1990, 1, 1, 0, 0, 0, 0).timestamp()


class DecodeError(Exception):
    """Decoding error exception object"""

    def __init__(
        self,
        format_index: int | None = None,
        format_name: str | None = None,
        letter_index: int | None = None,
        buffer_index: int | None = None,
        buffer_length: int | None = None,
        message: str | None = None,
    ):
        self.format_index = format_index
        self.format_name = format_name
        self.letter_index = letter_index
        self.buffer_index = buffer_index
        self.buffer_length = buffer_length
        self.message = message

    def as_dict(self) -> dict:
        return {
            "format_index": self.format_index,
            "format_name": self.format_name,
            "letter_index": self.letter_index,
            "buffer_index": self.buffer_index,
            "buffer_length": self.buffer_length,
            "message": self.message,
        }

    def __str__(self):
        return repr(self)

    def __repr__(self):
        variables_str = ", ".join(f"{k}={v}" for k, v in self.as_dict().items())

        return f"{self.__class__.__name__}({variables_str})"


def parse_gfp2(buffer: bytes) -> float:
    """Two-byte floating point decoder

    Parameters
    ----------
    buffer : bytes
        List of two values

    Returns
    -------
    float
        Decoded value
    """
    if len(buffer) < 2:
        raise ValueError("Buffer too short for gfp2 decoding")

    msb = buffer[0]
    lsb = buffer[1]
    Csign = -2 * (msb & 128) / 128 + 1
    CexpM = (msb & 64) / 64
    CexpL = (msb & 32) / 32
    Cexp = 2 * CexpM + CexpL - 3
    Cuppmant = (
        4096 * (msb & 16) / 16
        + 2048 * (msb & 8) / 8
        + 1024 * (msb & 4) / 4
        + 512 * (msb & 2) / 2
        + 256 * (msb & 1)
    )
    Cnum = Csign * (Cuppmant + lsb) * 10**Cexp
    return round(Cnum, 3)


def parse_gli4(buffer: bytes) -> int:
    """Four-byte decoder

    Parameters
    ----------
    buffer : bytes
        List of four values
        
    Returns
    -------
    float
        Decoded value
    """
    if len(buffer) < 4:
        raise ValueError("Buffer too short for gli4 decoding")

    return int.from_bytes(buffer[:4], byteorder="big", signed=True)


def determine_payload_format(payload: bytes, payload_formats_path: Path) -> str:
    """Determine payload format from lookup table, based on the first byte in
    the payload

    Parameters
    ----------
    payload : bytes
        Payload message
    payload_formats_path : Path
        File path to payload formats lookup table

    Returns
    -------
    bin_format : str
        Binary format of payload
    """
    payload_formats = pd.read_csv(payload_formats_path, index_col=0)
    payload_formats = payload_formats[payload_formats["flags"].values != "don’t use"]

    # The original code also had a special case for imei number 300234064121930. This should be handled differently.
    bidx = ord(payload[:1])

    logger.info(f"Format index: {bidx}")

    # The payload format is determined by the first byte of the payload
    if bidx not in payload_formats.index:
        raise DecodeError(format_index=bidx, message=f"Unknown payload format: {bidx}")

    payload_format = payload_formats.loc[bidx]
    _, bin_format, bin_name, flags, notes = payload_format.iloc[:5]

    logger.info(f"Format: {bin_name}, {notes}")
    logger.debug(f"Binary format: {bin_format}")

    return bin_format


def decode(bin_format: str, payload: bytes) -> list:
    """Decode a payload, based on a pre-defined format identifier

    Parameters
    ----------
    bin_format : str
        Binary payload format identifier
    payload : bytes
        Payload message

    Returns
    -------
    dataline : list
        Decoded payload
    """
    payload_length = len(payload)
    logger.info(f"Decoding payload with format: {bin_format!r}. Payload length: {payload_length}")
    logger.debug(f"Payload: {payload!r}")
    # Note: bin_val is just len(bin_format)
    indx = 1  # The first byte is the payload format
    dataline: list = []
    # Iterate over each token in the bin_format and decode the payload
    try:
        for format_letter_index, type_letter in enumerate(bin_format):
            logger.debug(
                f"Index {indx:02n} / {payload_length} Type letter: {type_letter:s}. upcuming bytes: {payload[indx:indx + 6]}..."
            )

            if type_letter == "f":
                # Encoded as 2 bytes base-10 floating point (GFP2)
                nan_values = (8191,)
                inf_values = (8190,)
                neg_inf_values = -8190, -8191

                value = parse_gfp2(payload[indx:])
                # Be careful with float and int comparison. Float doesn't guarantee exact values.
                # Consider using binary patterns for the special values instead.
                if value in nan_values:
                    dataline.append(np.nan)
                elif value in inf_values:
                    dataline.append(np.inf)
                elif value in neg_inf_values:
                    dataline.append(-np.inf)
                else:
                    dataline.append(value)
                indx += 2

            elif type_letter == "l":
                # Encoded as 4 bytes two complement integer (GLI4) - note the mac nan value is not a max value

                value = parse_gli4(payload[indx:])

                # The value 2147450879 seems arbitrary, but 2147483648 is the maximum value for a 4-byte signed integer
                nan_values = -2147483648, 2147450879
                if value in nan_values:
                    dataline.append(np.nan)
                else:
                    # TODO: Integers doesn't support NaN values and this will cause mixed types which should be avoided
                    dataline.append(float(value))
                indx += 4

            elif type_letter == "t":
                # timestamp as seconds since 1990-01-01 00:00:00 +0000 encoded as GLI4
                value = parse_gli4(payload[indx:])
                time = datetime.fromtimestamp(CR_BASIC_EPOCH_OFFSET + value)
                dataline.append(time)
                indx += 4

            elif type_letter in ("g", "n", "e"):
                # GPS time or coordinate encoding
                nan_values_fp2 = (8191,)
                # Check if byte is a 2-bit NAN. This occurs when the GPS data is not
                # available and the logger sends a 2-bytes NAN instead of a 4-bytes value
                if np.isnan(parse_gfp2(payload[indx:])) in nan_values_fp2:
                    # This is a 2-byte NAN
                    # The GPS data is not available
                    # We need to skip the next 2 bytes
                    dataline.append(np.nan)
                    indx += 2

                else:
                    # This is a 4-byte value
                    value = parse_gli4(payload[indx:])

                    if type_letter == "g":
                        value /= 100.0
                    else:  # type_letter in ('n', 'e'):
                        value /= 100000.0

                    dataline.append(value)
                    indx += 4
            else:
                raise Exception(f"Unknown type letter: {type_letter}")
            logger.debug(f"New value: {dataline[-1]}")

        # TODO: 2 bytes Floating points has overflow value 7999. Maybe also an underflow
        logger.debug(f"Index: {indx:02n} / {payload_length}")

        # Note: The CRBasic logger program us creating a checksum.
        # This is solely for the sending to the modem and it is not transmitted

        # TODO: Extract flags from suffix values
        # TODO: Investigate the protocol for the suffix values
        if indx != len(payload):
            raise Exception(f"Payload length mismatch: {indx} != {len(payload)}")

    except Exception as e:
        raise e
        # raise DecodeError(bidx, bin_name, format_letter_index, indx) from e
    return dataline


if __name__ == "__main__":
    import argparse
    import sys
    import pandas as pd
    from pathlib import Path

    parser = argparse.ArgumentParser(description="Payload decoder tool for CRBasic logger")
    parser.add_argument(
        "--payload_file_path",
        "-p",
        type=Path,
        help="Path to payload file",
        default=None,
    )
    parser.add_argument(
        "--format", "-f", help="Explicitly specify decoding string", default=None
    )
    parser.add_argument("--no-log", action="store_true", help="Disable logging", default=False)
    parser.add_argument(
        "--log_level",
        "-l",
        type=str,
        help="Set logging level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    )
    parser.add_argument(
        "--payload_formats_path",
        type=Path,
        help="Path to payload formats .csv file",
    )
    parser.add_argument(
        "--drop_checksum_suffix",
        action="store_true",
        help="Remove the last two bytes from the payload",
        default=False,
    )
    args = parser.parse_args()
    if args.no_log:
        logging.disable(logging.CRITICAL)
    else:
        logging.basicConfig(
            level=getattr(logging, args.log_level),
            format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            stream=sys.stderr
        )
        root_logger = logging.getLogger()
        root_logger.setLevel(getattr(logging, args.log_level))

    if isinstance(args.payload_file_path, Path):
        with open(args.payload_file_path, "rb") as payload_file:
            payload = payload_file.read()
    else:
        # Read payload from stdin
        payload = sys.stdin.buffer.read()

    if args.drop_checksum_suffix:
        payload = payload[:-2]

    if args.format is not None:
        payload_format = args.format
    else:
        if args.payload_formats_path is None:
            raise ValueError("Payload format path must be specified if decoding format is not specified")
        payload_format = determine_payload_format(payload, args.payload_formats_path)
    decoded = decode(payload_format, payload)

    df = pd.DataFrame([decoded]).to_csv(sys.stdout, index=False, header=False)
